JOB_TYPES_FOR_CONTENT_TYPE_VALIDATION = [
    "image_label_binary",
    "image_label_multiple_choice",
    "text_free_entry",
    "image_label_area_adjust",
    "image_label_area_select",
    "image_label_single_polygon",
    "image_label_multiple_polygons",
    "image_label_semantic_segmentation_one_option",
    "image_label_semantic_segmentation_multiple_options",
    "image_label_text",
]

SUPPORTED_CONTENT_TYPES = [
    "image/jpeg",
    "image/jpg",
    "image/png",
]
